@@ -1,5 +1,5 @@
 /* idna.c --- Prototypes for Internationalized Domain Name library.
-   Copyright (C) 2002-2015 Simon Josefsson
+   Copyright (C) 2002-2021 Simon Josefsson
 
    This file is part of GNU Libidn.
 
@@ -212,6 +212,11 @@ step3:
       }
     if (i < 64)
       out[i] = '\0';
+    else
+      {
+	free (src);
+	return IDNA_INVALID_LENGTH;
+      }
     if (inasciirange)
       goto step8;
   }
@@ -266,7 +271,7 @@ step3:
 
 step8:
   free (src);
-  if (strlen (out) < 1 || strlen (out) > 63)
+  if (strlen (out) < 1)
     return IDNA_INVALID_LENGTH;
 
   return IDNA_SUCCESS;
@@ -275,12 +280,12 @@ step8:
 /* ToUnicode().  May realloc() utf8in.  Will free utf8in unconditionally. */
 static int
 idna_to_unicode_internal (char *utf8in,
-			  uint32_t * out, size_t * outlen, int flags)
+			  uint32_t * out, size_t *outlen, int flags)
 {
   int rc;
   char tmpout[64];
   size_t utf8len = strlen (utf8in) + 1;
-  size_t addlen = 0;
+  size_t addlen = 0, addinc = utf8len / 10 + 1;
 
   /*
    * ToUnicode consists of the following steps:
@@ -320,7 +325,8 @@ idna_to_unicode_internal (char *utf8in,
 	rc = stringprep_nameprep (utf8in, utf8len + addlen);
       else
 	rc = stringprep_nameprep_no_unassigned (utf8in, utf8len + addlen);
-      addlen += 1;
+      addlen += addinc;
+      addinc *= 2;
     }
   while (rc == STRINGPREP_TOO_SMALL_BUFFER);
 
@@ -429,7 +435,7 @@ step3:
  */
 int
 idna_to_unicode_44i (const uint32_t * in, size_t inlen,
-		     uint32_t * out, size_t * outlen, int flags)
+		     uint32_t * out, size_t *outlen, int flags)
 {
   int rc;
   size_t outlensave = *outlen;
@@ -652,7 +658,10 @@ idna_to_unicode_4z4z (const uint32_t * input, uint32_t ** output, int flags)
       buflen = (size_t) (end - start);
       buf = malloc (sizeof (buf[0]) * (buflen + 1));
       if (!buf)
-	return IDNA_MALLOC_ERROR;
+	{
+	  free (out);
+	  return IDNA_MALLOC_ERROR;
+	}
 
       /* don't check return code as per specification! */
       idna_to_unicode_44i (start, (size_t) (end - start),
@@ -744,13 +753,16 @@ idna_to_unicode_8z8z (const char *input, char **output, int flags)
   int rc;
 
   rc = idna_to_unicode_8z4z (input, &ucs4, flags);
+  if (rc != IDNA_SUCCESS)
+    return rc;
+
   *output = stringprep_ucs4_to_utf8 (ucs4, -1, NULL, NULL);
   free (ucs4);
 
   if (!*output)
     return IDNA_ICONV_ERROR;
 
-  return rc;
+  return IDNA_SUCCESS;
 }
 
 /**
@@ -775,13 +787,16 @@ idna_to_unicode_8zlz (const char *input, char **output, int flags)
   int rc;
 
   rc = idna_to_unicode_8z8z (input, &utf8, flags);
+  if (rc != IDNA_SUCCESS)
+    return rc;
+
   *output = stringprep_utf8_to_locale (utf8);
   free (utf8);
 
   if (!*output)
     return IDNA_ICONV_ERROR;
 
-  return rc;
+  return IDNA_SUCCESS;
 }
 
 /**
@@ -843,7 +858,7 @@ idna_to_unicode_lzlz (const char *input, char **output, int flags)
  *   string does not equal the input.
  * @IDNA_CONTAINS_ACE_PREFIX: The input contains the ACE prefix (for
  *   ToASCII).
- * @IDNA_ICONV_ERROR: Could not convert string in locale encoding.
+ * @IDNA_ICONV_ERROR: Character encoding conversion error.
  * @IDNA_MALLOC_ERROR: Could not allocate buffer (this is typically a
  *   fatal error).
  * @IDNA_DLOPEN_ERROR: Could not dlopen the libcidn DSO (only used
